@@ -72,7 +72,7 @@ poly p_One(const ring r);
 
 int p_MinDeg(poly p,intvec *w, const ring R);
 
-long p_DegW(poly p, const short *w, const ring R);
+long p_DegW(poly p, const int *w, const ring R);
 
 /// return TRUE if all monoms have the same component
 BOOLEAN   p_OneComp(poly p, const ring r);
@@ -211,6 +211,7 @@ void      p_ProjectiveUnique(poly p,const ring r);
 
 void      p_ContentForGB(poly p, const ring r);
 void      p_Content(poly p, const ring r);
+void      p_Content_n(poly p, number &c,const ring r);
 #if 1
 // currently only used by Singular/janet
 void      p_SimpleContent(poly p, int s, const ring r);
@@ -676,16 +677,20 @@ static inline poly p_New(ring r)
   return p_New(r, r->PolyBin);
 }
 
-#if PDEBUG > 2
+#if (PDEBUG > 2) || defined(XALLOC_BIN)
 static inline void p_LmFree(poly p, ring r)
 #else
 static inline void p_LmFree(poly p, ring)
 #endif
 {
   p_LmCheckPolyRing2(p, r);
+  #ifdef XALLOC_BIN
+  omFreeBin(p,r->PolyBin);
+  #else
   omFreeBinAddr(p);
+  #endif
 }
-#if PDEBUG > 2
+#if (PDEBUG > 2) || defined(XALLOC_BIN)
 static inline void p_LmFree(poly *p, ring r)
 #else
 static inline void p_LmFree(poly *p, ring)
@@ -694,9 +699,13 @@ static inline void p_LmFree(poly *p, ring)
   p_LmCheckPolyRing2(*p, r);
   poly h = *p;
   *p = pNext(h);
+  #ifdef XALLOC_BIN
+  omFreeBin(h,r->PolyBin);
+  #else
   omFreeBinAddr(h);
+  #endif
 }
-#if PDEBUG > 2
+#if (PDEBUG > 2) || defined(XALLOC_BIN)
 static inline poly p_LmFreeAndNext(poly p, ring r)
 #else
 static inline poly p_LmFreeAndNext(poly p, ring)
@@ -704,14 +713,32 @@ static inline poly p_LmFreeAndNext(poly p, ring)
 {
   p_LmCheckPolyRing2(p, r);
   poly pnext = pNext(p);
+  #ifdef XALLOC_BIN
+  omFreeBin(p,r->PolyBin);
+  #else
   omFreeBinAddr(p);
+  #endif
   return pnext;
 }
 static inline void p_LmDelete(poly p, const ring r)
 {
   p_LmCheckPolyRing2(p, r);
   n_Delete(&pGetCoeff(p), r->cf);
+  #ifdef XALLOC_BIN
+  omFreeBin(p,r->PolyBin);
+  #else
   omFreeBinAddr(p);
+  #endif
+}
+static inline void p_LmDelete0(poly p, const ring r)
+{
+  p_LmCheckPolyRing2(p, r);
+  if (pGetCoeff(p)!=NULL) n_Delete(&pGetCoeff(p), r->cf);
+  #ifdef XALLOC_BIN
+  omFreeBin(p,r->PolyBin);
+  #else
+  omFreeBinAddr(p);
+  #endif
 }
 static inline void p_LmDelete(poly *p, const ring r)
 {
@@ -719,14 +746,22 @@ static inline void p_LmDelete(poly *p, const ring r)
   poly h = *p;
   *p = pNext(h);
   n_Delete(&pGetCoeff(h), r->cf);
+  #ifdef XALLOC_BIN
+  omFreeBin(h,r->PolyBin);
+  #else
   omFreeBinAddr(h);
+  #endif
 }
 static inline poly p_LmDeleteAndNext(poly p, const ring r)
 {
   p_LmCheckPolyRing2(p, r);
   poly pnext = pNext(p);
   n_Delete(&pGetCoeff(p), r->cf);
+  #ifdef XALLOC_BIN
+  omFreeBin(p,r->PolyBin);
+  #else
   omFreeBinAddr(p);
+  #endif
   return pnext;
 }
 
@@ -821,8 +856,8 @@ static inline poly p_Copy(poly p, const ring r)
     return NULL;
 }
 
-/// copy the i(leading) term of p
-static inline poly p_Head(poly p, const ring r)
+/// copy the (leading) term of p
+static inline poly p_Head(const poly p, const ring r)
 {
   if (p == NULL) return NULL;
   p_LmCheckPolyRing1(p, r);
@@ -835,8 +870,14 @@ static inline poly p_Head(poly p, const ring r)
   return np;
 }
 
+/// like p_Head, but allow NULL coeff
+poly p_Head0(const poly p, const ring r);
+
 /// like p_Head, but with coefficient 1
-poly p_CopyPowerProduct(poly p, const ring r);
+poly p_CopyPowerProduct(const poly p, const ring r);
+
+/// like p_Head, but with coefficient n
+poly p_CopyPowerProduct0(const poly p, const number n, const ring r);
 
 /// returns a copy of p with Lm(p) from lmRing and Tail(p) from tailRing
 static inline poly p_Copy(poly p, const ring lmRing, const ring tailRing)
@@ -1970,7 +2011,6 @@ static inline BOOLEAN p_IsConstantComp(const poly p, const ring r)
 static inline BOOLEAN p_IsConstant(const poly p, const ring r)
 {
   if (p == NULL) return TRUE;
-  p_Test(p, r);
   return (pNext(p)==NULL) && p_LmIsConstant(p, r);
 }
 
@@ -2096,8 +2136,8 @@ void p_SetModDeg(intvec *w, ring r);
 /*------------ Jet ----------------------------------*/
 poly pp_Jet(poly p, int m, const ring R);
 poly p_Jet(poly p, int m,const ring R);
-poly pp_JetW(poly p, int m, short *w, const ring R);
-poly p_JetW(poly p, int m, short *w, const ring R);
+poly pp_JetW(poly p, int m, int *w, const ring R);
+poly p_JetW(poly p, int m, int *w, const ring R);
 
 poly n_PermNumber(const number z, const int *par_perm, const int OldPar, const ring src, const ring dst);
 
